# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.xprophetnet.modeling import XProphetNetForGeneration
from unitorch.models.xprophetnet.processing import XProphetNetProcessor
